import asyncio
import sys
import argparse
from typing import AsyncGenerator, Literal
import ujson as json
from pathlib import Path
from vermils.asynctools import select
from asynctinydb import Query
from bookbaker import Config, Context, Book, Chapter, Episode, Task
from bookbaker.roles import DeepLTranslator, SyosetuComCrawler, GeminiTranslator
from bookbaker.roles import EpubExporter, AutoCrawler, GPTTranslator
from bookbaker.roles import BaseRole, BaseCrawler, BaseTranslator, BaseExporter
from bookbaker.roles import recover_from_dict


async def main():
    parser = argparse.ArgumentParser(
        prog="bookbaker",
        description="A light novel scraping and translating tool",
    )
    parser.add_argument(
        "action",
        type=str,
        choices=["crawl", "translate", "export", "all"],
        help=(
            "Action to perform\n"
            "crawl: Just crawl\n"
            "translate: Crawl and translate\n"
            "export: Just export\n"
            "all: Crawl, translate and export"
        ),
        default="all",
        nargs="?",
    )
    parser.add_argument("-c", "--config", type=str,
                        help="Path to config file", default="config.json")

    args = parser.parse_args()

    action: Literal["crawl", "translate", "export", "all"] = args.action
    config_path = Path(args.config)
    if not config_path.exists():
        print("Config file not found")
        user_input = input("Create a new config file? (Y/n): ").strip() or 'y'
        if user_input.lower() == 'y':
            config = Config()
            config.tasks.append(Task(
                url="https://syosetu.org/novel/333942/",
                friendly_name="sample - TS Tenshi",
                sauce_lang="JA",
                target_lang="ZH",
                crawler=None,
                translator=["gemini", "gpt"],
                exporter="epub",
                glossaries=[
                    ["ザラキエル", "萨拉琪尔"],
                    ["神専魔法", "神专魔法"],
                    ["サーヴィトリ", "萨维特丽"],
                    ["サラフィエル", "萨拉菲尔"],
                    ["アルマロス", "阿尔马洛斯"],
                    ["クロ", "小黑"]
                ]
            ))
            gmt = GeminiTranslator(name="gemini")
            gmt.backend.comp_tokens = 4096
            gmt.backend.temperature = 0.5
            config.roles.append(gmt)
            gpt = GPTTranslator(name="gpt")
            gpt.backend.comp_tokens = 2048
            gpt.backend.temperature = 0.4
            config.roles.append(gpt)
            config.roles.append(SyosetuComCrawler(name="syosetu-com"))
            config.roles.append(DeepLTranslator(name="deepl"))
            config.roles.append(EpubExporter(name="epub"))
            config_path.write_text(
                json.dumps(
                    config.model_dump(mode="json"),
                    indent=2,
                    ensure_ascii=False,
                ))
            print(f"Config file created at {config_path}")
            return 0
        else:
            return 2

    config = Config.model_validate_json(config_path.read_text())

    db = config.db.get_db()
    client = config.client.get_cli()
    logger = config.logging.get_logger()
    roles: list[BaseRole] = [recover_from_dict(d) for d in config.roles]

    def get_role(name: str):
        for role in roles:
            if role.name == name:
                return role
        raise ValueError(f"Role {name} not found")

    ctx = Context(
        db=db,
        client=client,
        logger=logger,
    )

    streams = list[AsyncGenerator[
        tuple[Task, Book, Chapter, Episode], None]]()

    async def crawl(crawler: BaseCrawler, t: Task, ctx: Context):
        async with t.lock:
            async for c in crawler.crawl_stream(t, ctx):
                t.lock.release()
                yield c
                await t.lock.acquire()

    if action in ("crawl", "translate", "all"):
        for t in config.tasks:
            crawler: BaseCrawler = AutoCrawler() if t.crawler is None else get_role(t.crawler)
            logger.info("Book %s (%s) added to crawl queue", t.friendly_name, t.url)
            streams.append(crawl(
                crawler,
                t,
                ctx=ctx
            ))

    async def translate(task: Task, book: Book, chapter: Chapter, episode: Episode):
        if (task.translator is None
                or task.translator == []
                or action not in ("translate", "all")):
            return
        async with task.lock:
            translators = list[BaseTranslator](
                [get_role(task.translator)]
                if isinstance(task.translator, str)
                else [get_role(t) for t in task.translator]
            )
            for translator in translators:
                logger.info("Translating %s with %s", episode.title, translator)

                try:
                    episode = await translator.translate(
                        episode,
                        task,
                        ctx,
                        chapter=chapter,
                        book=book,
                    )
                except Exception as e:
                    logger.exception(e)
                    if translator is translators[-1]:
                        logger.critical(
                            "Failed to translate %s with the last translator %s",
                            episode.title, translator)
                    else:
                        logger.warning("Failed to translate %s with %s",
                                       episode.title, translator)

    loop = asyncio.get_running_loop()
    aiotasks = list[asyncio.Task]()

    async for fut in select(streams, return_future=True):
        try:
            task, book, chapter, episode = fut.result()
        except Exception as e:
            logger.critical("Failed to crawl a url due to %s", e)
            logger.exception(e)
            continue

        aiotasks.append(loop.create_task(translate(task, book, chapter, episode)))

    await asyncio.gather(*aiotasks)

    if action in ("export", "all"):
        q = Query()
        for task in config.tasks:
            if task.exporter is None:
                continue
            exporters: list[BaseExporter] = (
                [get_role(task.exporter)]
                if isinstance(task.exporter, str)
                else [get_role(e) for e in task.exporter]
            )

            data = await ctx.db.get(q.url == task.url)
            if not data:
                logger.critical("No exporting data found for %s", task.url)
                continue
            book = Book.model_validate(data)

            for exporter in exporters:
                try:
                    logger.info("Exporting %s with %s", book.title, exporter)
                    data = await exporter.export(book, task, ctx)
                except Exception as e:
                    logger.critical("%s: Failed to export book %s",
                                    exporter, book.title)
                    logger.exception(e)

    logger.info("Closing...")
    await db.close()

sys.exit(asyncio.run(main()))



from __future__ import annotations

import re
from uuid import uuid4
from io import BytesIO
from pathlib import Path
from ebooklib import epub
from pydantic import Field
from base64 import b64encode
from vermils.io import aio
from ..classes import Book, Context, Task
from ..utils import get_url_content
from .base import BaseExporter


_a_img_matcher = re.compile(r"<a.*?href=\"(.+?)\".*?name=\"img\".*?>.*?<\/a>")
_img_matcher = re.compile(r"<img.*?src=\"(.*?)\".*?>")


class EpubExporter(BaseExporter):
    """
    Exporter for EPUB
    """
    name: str = Field(default_factory=lambda: f"epub-{str(uuid4())[:8]}")
    description: str = "EPUB Exporter"
    use_translated: bool = True

    async def export(
            self,
            book: Book,
            task: Task,
            ctx: Context,
    ) -> epub.EpubBook:
        logger = ctx.logger
        logger.info("%s: Exporting book %s", self, book.title)

        book_title = book.title_translated or book.title
        book_desc = book.description_translated or book.description
        lang = task.target_lang
        if self.use_translated:
            book_title = book.title_translated or book.title
            book_desc = book.description_translated or book.description
            lang = task.target_lang
        else:
            book_title = book.title
            book_desc = book.description
            lang = task.sauce_lang

        epub_book = epub.EpubBook()
        bid = str(hash((book.title, book.author)))
        epub_book.set_identifier(bid)
        epub_book.set_title(book_title)
        epub_book.set_language(lang)
        epub_book.add_author(book.author)

        if book.cover:
            cover_image: bytes | None = None
            try:
                cover_image = await get_url_content(book.cover.url, task.url, ctx)
                if cover_image:
                    epub_book.set_cover(
                        file_name=book.cover.alt or "cover", content=cover_image)
            except Exception as e:
                logger.warning(
                    "%s: Failed to set cover image for book: %s", self, book.title)
                logger.exception(e)

        for tag in book.tags:
            epub_book.add_metadata('DC', 'subject', tag)

        if book_desc:
            epub_chapter = epub.EpubHtml(title="Introduction", content=f"<p>{
                                         book_desc}</p>", file_name="intro.xhtml")
            epub_book.add_item(epub_chapter)
            epub_book.toc.append(epub_chapter)

        for chapter in book.chapters:
            if self.use_translated:
                chapter_title = chapter.title_translated or chapter.title
            else:
                chapter_title = chapter.title

            chapter_title = chapter_title or "MAIN_CHAPTER"
            epub_chapter = epub.EpubHtml(
                title=chapter_title,
                file_name=f"{chapter_title}.xhtml",
            )
            epub_chapter.content = f"<h1>{chapter_title}</h1>"

            logger.debug("%s: Adding chapter: %s to %s",
                         self, chapter_title, book_title)

            if chapter.cover:
                cover_image: bytes | None = None
                try:
                    cover_image = await get_url_content(chapter.cover.url, task.url, ctx)
                    epub_chapter.content += f"<img src=\"data:image/*;base64,{
                        b64encode(cover_image).decode()}\" />"
                except Exception as e:
                    logger.warning(
                        "%s: Failed to set cover image for chapter: %s", self, chapter.title)
                    logger.exception(e)
            epub_book.add_item(epub_chapter)
            epub_book.toc.append(epub_chapter)
            for episode in chapter.episodes:
                if self.use_translated:
                    episode_title = episode.title_translated or episode.title
                    episode_html = episode.html
                else:
                    episode_title = episode.title
                    episode_html = episode.raw_html
                epub_chapter = epub.EpubHtml(
                    title=episode_title, file_name=f"{episode_title}.xhtml")

                logger.debug("%s: Adding episode: %s to %s of %s",
                             self, episode_title, chapter_title, book_title)

                # Convert <a> tags pointing to images to <img> tags
                for m in _a_img_matcher.finditer(episode_html):
                    logger.debug("%s: Converting <a> tag to <img> tag: %s",
                                 self, m.group(0))
                    img_src = m.group(1)
                    episode_html = episode_html.replace(
                        m.group(0), f"<img src=\"{img_src}\" />")

                # Try to replace <img> tags with base64 encoded images
                for m in _img_matcher.finditer(episode_html):
                    img_src = m.group(1)
                    try:
                        logger.debug(
                            "%s: Converting image to base64: %s", self, img_src)
                        img_content = await get_url_content(img_src, task.url, ctx)
                        img_content = b64encode(img_content)
                        img_src = f"data:image/*;base64,{img_content.decode()}"
                        episode_html = episode_html.replace(
                            m.group(0), f"<img src=\"{img_src}\" />")
                    except Exception as e:
                        logger.warning(
                            "%s: Failed to convert image to base64: %s", self, img_src)
                        logger.exception(e)

                epub_chapter.content = episode_html
                epub_book.add_item(epub_chapter)
                epub_book.toc.append(epub_chapter)

        epub_book.add_item(epub.EpubNcx())
        epub_book.add_item(epub.EpubNav())
        epub_book.spine = ['nav'] + epub_book.toc

        if not await aio.path.exists("exports"):
            await aio.os.mkdir("exports")
        path = "exports" / Path(f"{book.title}({lang}).epub")

        ebook_bytes = BytesIO()
        epub.write_epub(ebook_bytes, epub_book, {})
        ebook_bytes.seek(0)

        async with aio.open(str(path), "wb") as f:
            await f.write(ebook_bytes.getvalue())

        return epub_book



from __future__ import annotations

import bs4
import httpx
from typing import cast
from uuid import uuid4
from pydantic import Field
from asynctinydb import Query, Document
from datetime import datetime, UTC
from urllib.parse import urlparse
from .base import BaseCrawler
from ..classes import Book, Chapter, Episode, Line, TimeMeta, Context, Task


class SyosetuOrgCrawler(BaseCrawler):
    """
    Crawler for syosetu.org
    """
    name: str = Field(default_factory=lambda: f"syosetu-org-{str(uuid4())[:8]}")
    description: str = "A crawler for syosetu.org"

    async def crawl_stream(
            self,
            task: Task,
            ctx: Context,
    ):
        logger = ctx.logger
        db = ctx.db
        cli = ctx.client
        url = urlparse(task.url)
        logger.info("%s: Crawling URL %s", self, task.url)
        r = await cli.get(task.url)
        r.raise_for_status()

        soup = bs4.BeautifulSoup(r.text.replace('\u3000', "  "), "lxml")
        body = soup.body
        title = body.find(itemprop="name").text.strip()
        author = body.find(itemprop="author").text.strip()
        logger.info("%s: Crawling book %s by %s", self, title, author)

        book_query = Query()
        data: Document = await db.get(
            book_query.title == title and book_query.author == author)
        if data:
            logger.debug("%s: Book %s retrieved from database", self, title)
            book = Book.model_validate(dict(data))
        else:
            book = Book(title=title, author=author)

        book.url = task.url

        tags = set[str]()
        genre = body.find(itemprop="genre")
        for child in genre.children:
            if not isinstance(child, bs4.Tag):
                continue
            tags.add(child.text.strip())
        for tag in body.find_all('a', class_="alert_color"):
            tag: bs4.Tag
            tags.add(tag.text.strip())
        for tag in body.find_all(itemprop="keywords"):
            tags.add(tag.text.strip())
        book.tags.update(tags)

        maind = body.find(id="maind")
        divs = maind.find_all("div")
        desc_div = divs[2]
        desc = ''
        for br in desc_div.find_all('br'):
            br.replace_with(f"\n{br.text}")
        for child in desc_div.children:
            if isinstance(child, bs4.NavigableString):
                desc += str(child)
        book.description = desc

        indexes = maind.find("table")
        default_chapter = book.get_chapter('')

        if default_chapter is None:
            default_chapter = Chapter(title='')
            book.chapters.append(default_chapter)
        chapter = default_chapter

        for child in indexes.find_all("tr"):
            child: bs4.Tag
            if len(child.find_all("td")) == 1:
                chapter_name = child.text.strip()
                logger.info("%s: Crawling chapter %s", self, chapter_name)
                old_chapter = book.get_chapter(chapter_name)
                if old_chapter is None:
                    logger.debug("%s: New chapter %s created", self, chapter_name)
                    chapter = Chapter(title=chapter_name)
                    book.chapters.append(chapter)
                else:
                    logger.debug("%s: Chapter %s retrieved from database",
                                 self, chapter_name)
                    chapter = old_chapter
            else:
                tds = list[bs4.Tag](child.find_all("td"))
                episode_url = cast(str, tds[0].a["href"])
                if episode_url.startswith("./"):
                    path = url.path.removesuffix("/")
                    path += episode_url[1:]
                    episode_url = f"{url.scheme}://{url.netloc}{path}"

                subtitle = tds[0].text.strip()
                logger.info("%s: Crawling episode %s from %s",
                            self, subtitle, episode_url)

                created_at = self._parse_datatime(
                    tds[1].text.strip()
                )
                updates = tds[1].find_all("span")
                updated_at = self._parse_datatime(
                    updates[-1]["title"]) if updates else created_at

                episode = chapter.get_episode(subtitle)
                if episode is not None:
                    logger.debug("%s: Episode %s retrieved from database",
                                 self, subtitle)
                    if episode.time_meta.created_at is None:
                        episode.time_meta.created_at = created_at
                    episode.time_meta.updated_at = updated_at
                    created_at = episode.time_meta.created_at

                    if episode.time_meta.saved_at < updated_at:
                        logger.debug("%s: Episode %s updated", self, subtitle)
                        episode_index = chapter.episodes.index(episode)
                        episode = await self.get_episode(
                            episode_url, cli,
                            created_at=created_at, updated_at=updated_at)
                        chapter.episodes[episode_index] = episode
                else:
                    logger.debug("%s: New episode %s created", self, subtitle)
                    episode = await self.get_episode(
                        episode_url, cli,
                        created_at=created_at, updated_at=updated_at)
                    chapter.episodes.append(episode)

                if not episode.lines:
                    logger.warning("%s: Episode %s has no content", self, subtitle)

                await db.upsert(book.model_dump(mode="json"),
                                book_query.title == title
                                and book_query.author == author)

                yield task, book, chapter, episode

        if not default_chapter.episodes:
            book.chapters.remove(default_chapter)

        await db.upsert(book.model_dump(mode="json"),
                        book_query.title == title and book_query.author == author)

    def _parse_datatime(self, dt: str) -> datetime:
        """e.g. '2024年01月16日(火) 22:08改稿'"""
        try:
            dt = dt.strip()
            year, month, day, hour, minute = map(int, (
                dt[:4], dt[5:7], dt[8:10], dt[15:17], dt[18:20])
            )
            return datetime(
                year=year, month=month, day=day, hour=hour, minute=minute
            ).astimezone(UTC)
        except Exception as e:
            raise ValueError(f"Failed to parse datetime: {dt}") from e

    async def get_episode(
            self,
            url: str,
            client: httpx.AsyncClient,
            *,
            created_at: datetime | None = None,
            updated_at: datetime | None = None,
    ) -> Episode:
        r = await client.get(url)
        r.raise_for_status()

        if created_at is not None:
            created_at = created_at.astimezone(UTC)
        if updated_at is not None:
            updated_at = updated_at.astimezone(UTC)

        soup = bs4.BeautifulSoup(r.text.replace('\u3000', "  "), "lxml")
        body = soup.body
        title = soup.find(
            "meta", property="og:title")["content"].split(" - ", maxsplit=2)[0]

        honbun = body.find(id="honbun")
        lines = list[Line]()
        for p in honbun.find_all("p"):
            p: bs4.Tag
            for br in p.find_all("br"):
                br: bs4.Tag
                br.replace_with(f"\n{br.decode_contents()}")
            decoded = str(p.decode_contents())
            if not decoded.strip():
                decoded = ''  # replace empty lines with empty string
            lines.append(Line(content=decoded))
        episode = Episode(
            title=title,
            lines=lines,
            time_meta=TimeMeta(
                created_at=created_at, updated_at=updated_at),
        )
        preview = body.find(id="maegaki")
        if preview:
            for br in preview.find_all("br"):
                br: bs4.Tag
                br.replace_with(f"\n{br.decode_contents()}")
            episode.notes = preview.text

        return episode



from __future__ import annotations

from uuid import uuid4
from pydantic import Field
from urllib.parse import urlparse
from .base import BaseCrawler
from .syosetu_com import SyosetuComCrawler
from .syosetu_org import SyosetuOrgCrawler
from .kakuyomu import KakuyomuCrawler
from .novelup import NovelUpCrawler
from ..classes import Context, Task


class AutoCrawler(BaseCrawler):
    """
    Automatically choose the right crawler for the given URL
    """
    name: str = Field(default_factory=lambda: f"auto-{str(uuid4())[:8]}")
    description: str = "Automatically choose crawler to use"

    def crawl_stream(
            self,
            task: Task,
            ctx: Context,
    ):
        url = urlparse(task.url)
        hostname = url.hostname
        crawler: BaseCrawler
        if hostname is None:
            raise ValueError(f"Invalid URL: {task.url}")
        if hostname.endswith("syosetu.com"):
            crawler = SyosetuComCrawler()
        elif hostname.endswith("syosetu.org"):
            crawler = SyosetuOrgCrawler()
        elif hostname.endswith("kakuyomu.jp"):
            crawler = KakuyomuCrawler()
        elif hostname.endswith("novelup.plus"):
            crawler = NovelUpCrawler()
        else:
            raise ValueError(f"Cannot determine the crawler for {task.url}")

        return crawler.crawl_stream(task, ctx)



from __future__ import annotations

import re
import bs4
import httpx
from typing import cast
from uuid import uuid4
from pydantic import Field
from asynctinydb import Query, Document
from datetime import datetime, UTC
from urllib.parse import urlparse
from .base import BaseCrawler
from ..classes import Book, Chapter, Episode, Line, TimeMeta, Context, Task


class NovelUpCrawler(BaseCrawler):
    """
    Crawler for novelup.plus
    """
    name: str = Field(default_factory=lambda: f"novelup-{str(uuid4())[:8]}")
    description: str = "A crawler for novelup.plus"

    _dt_re = re.compile(r".*?(\d+)\D(\d{1,2})\D(\d{1,2})\D+(\d{1,2})\D(\d{1,2}).*")
    _ep_ord_re = re.compile(r"^(\d+)")

    async def crawl_stream(
            self,
            task: Task,
            ctx: Context,
    ):
        logger = ctx.logger
        db = ctx.db
        cli = ctx.client
        url = urlparse(task.url)
        origin = f"{url.scheme}://{url.netloc}"
        logger.info("%s: Crawling URL %s", self, task.url)
        r = await cli.get(task.url)
        r.raise_for_status()

        soup = bs4.BeautifulSoup(r.text.replace('\u3000', "  "), "xml")
        body = soup.body
        title = body.find(class_="novel_title").text.strip()
        author = body.find(class_="novel_author").a.text.strip()
        logger.info("%s:  book %s by %s", self, title, author)

        book_query = Query()
        data: Document = await db.get(
            book_query.title == title and book_query.author == author)
        if data:
            logger.debug("%s: Book %s retrieved from database", self, title)
            book = Book.model_validate(dict(data))
        else:
            book = Book(title=title, author=author)

        book.url = task.url

        desc = body.find(class_="novel_synopsis")
        for br in desc.find_all('br'):
            br.replace_with(f"\n{br.text}")
        book.description = desc.text.strip()

        info_div = body.find(id="section_episode_info_table")
        info_tab = info_div.find("table")
        for tr in info_tab.find_all("tr"):
            if "ジャンル" in tr.th.text:
                genres = tr.td.text.strip().split()
                book.tags.update(genres)
            elif "タグ" in tr.th.text:
                tags = tr.td.text.strip().split()
                book.tags.update(tags)
            elif "セルフレイティング" in tr.th.text:
                ratings = tr.td.text
                if "残酷描写" in ratings:
                    book.tags.add("残酷描写")
                if "暴力描写" in ratings:
                    book.tags.add("暴力描写")
                if "性的表現" in ratings:
                    book.tags.add("性的表現")
            elif "初掲載日" in tr.th.text:
                book.time_meta.created_at = self._parse_datatime(tr.td.text)
            elif "最終更新日" in tr.th.text:
                book.time_meta.updated_at = self._parse_datatime(tr.td.text)

        indexes = body.find(class_="episode_list")
        default_chapter = book.get_chapter('')

        if default_chapter is None:
            default_chapter = Chapter(title='')
            book.chapters.append(default_chapter)
        chapter = default_chapter

        for child in indexes.find_all("li"):
            if not isinstance(child, bs4.Tag):
                continue
            if "chapter" in child["class"]:
                chapter_name = child.text.strip()
                logger.info("%s: Crawling chapter %s", self, chapter_name)
                old_chapter = book.get_chapter(chapter_name)
                if old_chapter is None:
                    logger.debug("%s: New chapter %s created", self, chapter_name)
                    chapter = Chapter(title=chapter_name)
                    book.chapters.append(chapter)
                else:
                    logger.debug("%s: Chapter %s retrieved from database",
                                 self, chapter_name)
                    chapter = old_chapter
            else:
                div = child.find("div")
                episode_url = cast(str, div.a["href"])
                if episode_url.startswith("/"):
                    episode_url = f"{origin}{episode_url}"

                subtitle = div.text.strip()
                subtitle = self._ep_ord_re.sub(
                    '', subtitle).strip()  # remove episode number
                logger.info("%s: Crawling episode %s from %s",
                            self, subtitle, episode_url)

                dates = child.find(class_="update_date")
                created_at = self._parse_datatime(
                    dates.find("span").text.replace('\n', ' '))
                updated_at = created_at
                next_span = dates.find_next_sibling("span")
                if next_span:
                    updated_at = self._parse_datatime(
                        next_span.text.replace('\n', ' '))

                episode = chapter.get_episode(subtitle)
                if episode is not None:
                    logger.debug("%s: Episode %s retrieved from database",
                                 self, subtitle)
                    if episode.time_meta.created_at is None:
                        episode.time_meta.created_at = created_at
                    episode.time_meta.updated_at = updated_at
                    created_at = episode.time_meta.created_at

                    if episode.time_meta.saved_at < updated_at:
                        logger.debug("%s: Episode %s updated", self, subtitle)
                        episode_index = chapter.episodes.index(episode)
                        episode = await self.get_episode(
                            episode_url, cli,
                            created_at=created_at, updated_at=updated_at)
                        chapter.episodes[episode_index] = episode
                else:
                    logger.debug("%s: New episode %s created", self, subtitle)
                    episode = await self.get_episode(
                        episode_url, cli,
                        created_at=created_at, updated_at=updated_at)
                    chapter.episodes.append(episode)

                if not episode.lines:
                    logger.warning("%s: Episode %s has no content", self, subtitle)

                await db.upsert(book.model_dump(mode="json"),
                                book_query.title == title and book_query.author == author)

                yield task, book, chapter, episode

        if not default_chapter.episodes:
            book.chapters.remove(default_chapter)

        await db.upsert(book.model_dump(mode="json"),
                        book_query.title == title and book_query.author == author)

    def _parse_datatime(self, dt: str) -> datetime:
        try:
            dts = self._dt_re.match(dt)
            if dts is None:
                raise ValueError(f"Failed to parse datetime: {dt}")
            year, month, day, hour, minute = map(int, dts.groups())
            return datetime(
                year=year, month=month, day=day, hour=hour, minute=minute
            ).astimezone(UTC)
        except Exception as e:
            raise ValueError(f"Failed to parse datetime: {dt}") from e

    async def get_episode(
            self,
            url: str,
            client: httpx.AsyncClient,
            *,
            created_at: datetime | None = None,
            updated_at: datetime | None = None,
    ) -> Episode:
        r = await client.get(url, cookies={"over18": "yes"})
        r.raise_for_status()

        if created_at is not None:
            created_at = created_at.astimezone(UTC)
        if updated_at is not None:
            updated_at = updated_at.astimezone(UTC)

        soup = bs4.BeautifulSoup(r.text.replace('\u3000', "  "), "xml")
        body = soup.body
        title = body.find(class_="episode_title").text.strip()
        honbun = body.find(id="episode_content")
        for br in honbun.find_all("br"):
            br: bs4.Tag
            br.replace_with(f"\n{br.decode_contents()}")

        lines = [Line(content=p) for p in honbun.text.splitlines()]

        episode = Episode(
            title=title,
            lines=lines,
            time_meta=TimeMeta(
                created_at=created_at, updated_at=updated_at),
        )
        preview = body.find(class_="novel_afterword")
        if preview:
            episode.notes = preview.text

        return episode



from __future__ import annotations

import ujson as json
import asyncio
from typing import overload
from uuid import uuid4
from pydantic import Field
from asynctinydb import Query
from gemnine import Bot, Message, Role
from ..misc import LANG_NAME_TABLE
from ..utils import escape_ruby, unescape_ruby
from ..classes import Context, Task
from ..classes import Book, Chapter, Episode
from .base import BaseTranslator


class GeminiTranslator(BaseTranslator):
    """
    # GeminiTranslator Class
    """
    name: str = Field(default_factory=lambda: f"gemini-{str(uuid4())[:8]}")
    description: str = "A translator for gemini.com"
    max_retries: int | None = 10
    """Max retries before giving up"""
    batch_size: int = 1024
    """Max characters to send in one batch"""
    max_tokens: int | None = None
    """Max tokens to preserve"""
    remind_interval: int | None = 3
    """Interval to remind glossaries"""
    skip_translated: bool = True
    """Skip already translated lines"""
    convert_ruby: bool = True
    """Convert <ruby> tags to simpler format for LLMs"""
    backend: Bot = Field(
        default_factory=lambda: Bot(model="models/gemini-pro", api_key=''))

    async def translate(
            self,
            episode: Episode,
            task: Task,
            ctx: Context,
            chapter: Chapter | None = None,
            book: Book | None = None,
    ) -> Episode:

        if episode.fully_translated and self.skip_translated:
            return episode

        cli = ctx.client
        self.backend._cli = cli
        db = ctx.db
        sauce_lang = LANG_NAME_TABLE.get(task.sauce_lang.upper(), task.sauce_lang)
        target_lang = LANG_NAME_TABLE.get(task.target_lang.upper(), task.target_lang)
        logger = ctx.logger
        sess = task.extra.get(self.name, self.backend.new_session())
        task.extra[self.name] = sess
        sess.message_lock = 2  # Prevents first 2 prompts being deleted

        prompt = (
            "You're a professional translator, "
            f"converting text from {sauce_lang} to {target_lang}. "
            "Correctly add missing subject and follow references. "
            f"Rephrase for {target_lang} naturalness and correct errors. "
            "For JSON, maintain keys with translated values. "
            "Keep line count for pure text and preserve '\\n' and [](^) if present. "
            "Good translation examples for Japanese to Chinese:\n"
            "西に傾きかかった太陽は、この丘の裾遠く広がった有明の入り江の上に、長く曲折しつつはるか水平線の両端に消え入る白い沙丘の上に今は力なく其の光を投げていた。\n"
            "西斜的太阳，无力地照射着山脚下向远处扩展的有明海海湾，照射着蜿蜒曲折地消失在海平线远方的白色沙丘。\n"
            "事故が起こらないように、十分運転にお気を付けてください。\n"
            "请注意安全驾驶，以免发生事故。\n"
            "事件の詳しい経過がわかり次第、番組の中でお手伝えいたします。\n"
            "一旦弄清事情的详细经过，我们将随时在节目中报道。\n"
            "並んでいるね。\n"
            "这么多人排队啊。\n"
            "彼の言うことは、あるいは本当かもしれない。\n"
            "他说的或许是真的。\n"
            "「きみ、あたまいいね。」「よく言われるんだよ。」\n"
            "“你很聪明嘛！”“大家都这么说。”\n"
            "周囲を完全に包囲したから、犯人はもう袋のねずみだ。\n"
            "四下里都包围得如铁桶一般，这下他可是插翅难飞了。\n"
            "我々は金大郎飴のように同じような車を作るつもりはない。\n"
            "我们不打算生产过于雷同的车型。\n"
            "「いつか作家になるか、起業して楽しく暮らす」という夢を持っていた。\n"
            "我曾有个梦想，当个作家，或是自己做个小生意，过上幸福的生活。\n"
        )
        ex_prompt = prompt
        if task.glossaries:
            ex_prompt += "Translation reference you must follow:\n"
            ex_prompt += '\n'.join(f"{k} : {v}" for k, v in task.glossaries)
        if book is not None:
            ex_prompt += "\nThe book you are translating:\n"
            book_full_meta = {
                "title": book.title,
                "description": book.description,
                "series": book.series,
                "tags": list(book.tags),
            }
            ex_prompt += f"{json.dumps(book_full_meta, ensure_ascii=False)}\n"

        ex_prompt += "\nThe episode you are translating:\n"
        episode_meta: dict[str, str | None] = {
            "title": episode.title,
            "notes": episode.notes,
        }
        ex_prompt += f"{json.dumps(episode_meta, ensure_ascii=False)}\n"
        ex_prompt += (
            "\nYou can start translating now: {\"test_en\": \"りんごはおいしい！\", \"test_zh\": \"りんごはおいしい！\"")

        logger.debug("%s: generated prompt: %s", self, ex_prompt)

        sess.append(Message(role=Role.User, parts=ex_prompt))
        sess.append(Message(role=Role.Model,
                    parts="{\"test_en\": \"Apple is yummy!, \"test_zh\": \"苹果很好吃！\""))

        def remind():
            logger.debug("%s: Sending reminder", self)
            sess.append(
                Message(role=Role.User, parts=f"{prompt}\nTranslation references [{", ".join(g[0] for g in task.glossaries)}]"))
            sess.append(
                Message(role=Role.Model, parts=f"[{", ".join(g[1] for g in task.glossaries)}]"))
        remind()
        cycle = 0

        @overload
        async def translate(c: dict[str, str | None]) -> dict[str, str | None]:
            ...

        @overload
        async def translate(c: list[str]) -> list[str]:
            ...

        async def translate(
                c: list[str] | dict[str, str | None]
        ) -> list[str] | dict[str, str | None]:
            nonlocal cycle
            if isinstance(c, dict):
                pstr = json.dumps(c, ensure_ascii=False)
            else:
                pstr = '\n'.join(map(lambda x: x.replace('\n', r"\n"), c))

            if self.convert_ruby:
                pstr = escape_ruby(pstr)

            logger.debug("%s: Sending prompt: %s", self, pstr)
            await sess.trim(self.max_tokens)
            sess_bak = sess.messages.copy()

            retry = 0

            while True:
                resp: str = "None"
                try:
                    resp = await sess.send(pstr)

                    if self.convert_ruby:
                        resp = unescape_ruby(resp)

                    logger.debug("%s: Received response: %s", self, resp)
                    if isinstance(c, list):
                        obj = resp.split('\n')
                        obj = list(
                            filter(None, map(lambda x: x.replace(r"\n", '\n'), obj)))
                        if not isinstance(obj, list):
                            raise TypeError("Expected list, got dict")
                        if len(obj) != len(c):
                            raise ValueError("Length mismatch")
                    else:
                        obj = json.loads(resp)
                        if not isinstance(obj, dict):
                            raise TypeError("Expected dict, got list")
                        if obj.keys() != c.keys():
                            raise ValueError("Key mismatch")

                    return obj

                except Exception as e:
                    logger.debug("%s: Failed to get valid response: %s", self, resp)
                    logger.exception(e)
                    sess.messages = sess_bak
                    retry += 1
                    if self.max_retries is not None and retry > self.max_retries:
                        raise RuntimeError(
                            "Failed to get valid response in %d retries" % self.max_retries)
                    await asyncio.sleep(1)
                    continue

                finally:
                    cycle += 1
                    if self.remind_interval is not None and cycle >= self.remind_interval:
                        remind()
                        cycle = 0

        if book is not None and None in (
            book.title_translated,
            book.description_translated,
            book.series_translated if book.series else ''
        ):
            book_meta: dict[str, str | None] = {
                "title": book.title,
                "description": book.description,
                "series": book.series
            }
            book_meta = await translate(book_meta)
            book.title_translated = book_meta["title"]
            book.description_translated = book_meta["description"]
            book.series_translated = book_meta["series"]

        if chapter is not None and chapter.title_translated is None:
            chapter_meta: dict[str, str | None] = {
                "title": chapter.title
            }
            chapter_meta = await translate(chapter_meta)
            chapter.title_translated = chapter_meta["title"]

        episode_meta = await translate(episode_meta)
        episode.title_translated = episode_meta["title"]
        episode.notes_translated = episode_meta["notes"]

        indexes = list[int]()
        contents = list[str]()
        c_cnt = 0
        MAX_CNT = self.batch_size

        for i, line in enumerate(episode.lines):
            if not line.content.strip():
                continue
            if self.skip_translated and line.translated is not None:
                continue
            indexes.append(i)
            contents.append(line.content)
            c_cnt += len(line.content)
            if c_cnt > MAX_CNT:
                translated = await translate(contents)
                for j, v in zip(indexes, translated):
                    o = episode.lines[j].content
                    line = episode.lines[j]
                    line.translated = v
                    line.candidates[self.name] = v
                    if not v.strip():
                        logger.warning("%s: Empty translation for %s", self, o)
                indexes.clear()
                contents.clear()
                c_cnt = 0

        if indexes:
            translated = await translate(contents)
            for j, v in zip(indexes, translated):
                line = episode.lines[j]
                line.translated = v
                line.candidates[self.name] = v
                if not v.strip():
                    logger.warning("%s: Empty translation for %s", self, o)

        q = Query()
        if book is not None:
            await db.upsert(book.model_dump(mode="json"),
                            q.title == book.title and q.author == book.author)

        return episode
